"""
Helper functions for performing coord check.
"""
from copy import copy

import pandas as pd
import tensorflow as tf

from mup_tf.layer import extract_layers


def cov(x):
    """Treat `x` as a collection of vectors and its Gram matrix.

    Args:
        x: A tf.tensor that if it has shape [..., d], then it's treated as
            a collection of d-dimensional vectors
    Returns:
        a matrix of size N x N where N is the product of
            the non-last dimensions of `x`.
    """
    if tf.size(x) == 1:
        width = 1
        xx = tf.reshape(x, (1, 1))
    else:
        width = x.shape[-1]
        xx = tf.reshape(x, (-1, width))
    return xx @ tf.transpose(xx) / width


def covoffdiag(x):
    """Get off-diagonal entries of `cov(x)` in a vector.

    Args:
        x: A tf.tensor it has shape [..., d], then it's treated as
            a collection of d-dimensional vectors
    Returns:
        Off-diagonal entries of `cov(x)` in a vector."""
    c = cov(x)
    return c[~tf.eye(c.shape[0], dtype=bool)]


#: dict of provided functions for use in coord check
FDICT = {
    "l1": lambda x: tf.math.reduce_mean(tf.math.abs(x)),
    "l2": lambda x: tf.math.reduce_mean(x**2) ** 0.5,
    "mean": lambda x: tf.math.reduce_mean(x),
    "std": lambda x: tf.math.reduce_std(x),
    "covl1": lambda x: tf.math.reduce_mean(tf.math.abs(cov(x))),
    "covl2": lambda x: tf.math.reduce_mean(cov(x) ** 2) ** 0.5,
    "covoffdiagl1": lambda x: tf.math.reduce_mean(tf.math.abs(covoffdiag(x))),
    "covoffdiagl2": lambda x: tf.math.reduce_mean((covoffdiag(x) ** 2)) ** 0.5,
}


def convert_fdict(d):
    """convert a dict `d` with string values to function values.

    Args:
        d: a dict whose values are either strings or functions
    Returns:
        a new dict, with the same keys as `d`, but the string values are
        converted to functions using `FDICT`.
    """
    return dict(
        [((k, FDICT[v]) if isinstance(v, str) else (k, v)) for k, v in d.items()]
    )


def get_input_output(model, data, t, width, output_fdict=None, skip_tensors=None):
    """
    Args:
        model: a tf.keras.models.Model
        data: a batch of data
        t: int representing the time step
        width: int representing the width of the model
        output_fdict: dict of functions to apply to the output of the model

    Returns:
        a list of dicts where a function in `output_fdict` is applied to the
        output of each layer of the model.
    """
    if output_fdict is None:
        output_fdict = dict(l1=FDICT["l1"])
    else:
        output_fdict = convert_fdict(output_fdict)

    records = []

    output = data
    layers = extract_layers(model, extract_custom=False)

    for i, tensor in enumerate(layers):
        if skip_tensors and skip_tensors in tensor.name:
            continue

        output = tensor(output, training=False)
        name = f"{tensor.name.split('_')[0]}_{i}"
        ret = {"width": width, "module": name, "t": t}

        for fname, f in output_fdict.items():
            ret[fname] = f(output).numpy()

        records.append(ret)

    return records


def _get_coord_data(
    models,
    dataloader,
    optimizer,
    nsteps=3,
    dict_in_out=False,
    dict_out=False,
    flatten_input=False,
    flatten_output=False,
    output_name="loss",
    lossfn="xent",
    filter_module_by_name=None,
    fix_data=True,
    nseeds=1,
    output_fdict=None,
    input_fdict=None,
    param_fdict=None,
    show_progress=True,
    one_hot_target=False,
    skip_tensors=None,
):
    """Inner method for `get_coord_data`.

    Train the models in `models` with optimizer given by `optimizer` and data from
    `dataloader` for `nsteps` steps, and record coordinate statistics specified
    by `output_fdict`, `input_fdict`, `param_fdict`. By default, only `l1` is
    computed for output activations of each module.

    Args:
        models:
            a dict of lazy models, where the keys are numbers indicating width.
            Each entry of `models` is a function that instantiates a model given
            nothing.
        dataloader:
            an iterator whose elements are either Huggingface style dicts, if
            `dict_in_out` is True, or (input, label). If `fix_data` is True
            (which is the default), then only the first element of `dataloader`
            is used in a loop and the rest of `dataloder` is ignored.
        optimizer:
            a tf.keras.optimizers.Optimizer or an optimizer implemented in `mup_tf`.
        nsteps:
            number of steps to train the model
        dict_in_out:
            whether the data loader contains Huggingface-style dict input and
            output. Default: False
        flatten_input:
            if not `dict_in_out`, reshape the input to be
            `tf.reshape(input, (input.shape[0], -1)`. Typically used for testing MLPs.
        flatten_output:
            if not `dict_in_out`, reshape the label to be `tf.reshape(label, (-1, label.shape[-1]))`.
        output_name:
            if `dict_in_out`, this is the key for the loss value if the output
            is a dict. If the output is not a dict, then we assume the first
            element of the output is the loss.
        lossfn:
            loss function to use if not `dict_in_out`. Can be either a string from
            [`xent`, 'mse'] Default: 'xent'
        filter_module_by_name:
            a function that returns a bool given layer names (from
            `model.layers`), or None. If not None, then only modules
            whose name yields True will be recorded.
        nseeds:
            number of times to repeat the training, each with different seeds.
        output_fdict, input_fdict, param_fdict:
            function dicts to be used in `_record_coords`. By default, only `l1`
            is computed for output activations of each module.
        show_progress:
            show progress using tqdm. Default: True
        one_hot_target:
            convert target label into a one-hot vector. This typically is only
            used for `'mse'` or `'l1'` losses in classification tasks.
            Default: False
    Returns:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'module', 't'` as well as names of statistics recorded, such
        as `'l1'` (see `FDICT` for other premade statistics that can be
        collected).
    """
    df = []
    if fix_data:
        batch = next(iter(dataloader))
        dataloader = [batch] * nsteps
    if show_progress:
        from tqdm import tqdm

        pbar = tqdm(total=nseeds * len(models))

    for i in range(nseeds):
        tf.random.set_seed(i)
        for width, model in models.items():
            model = model()
            for t, batch in enumerate(dataloader, 1):
                with tf.GradientTape() as tape:
                    if dict_in_out:
                        outputs = model(**batch, training=True)
                        output = (
                            outputs[output_name]
                            if isinstance(outputs, dict)
                            else outputs[0]
                        )

                    elif dict_out:
                        (data, target) = batch
                        outputs = model(data, training=True)
                        output = (
                            outputs[output_name]
                            if isinstance(outputs, dict)
                            else outputs[0]
                        )

                    else:
                        (data, target) = batch
                        if flatten_input:
                            data = tf.reshape(data, (data.shape[0], -1))
                        output = model(data, training=True)
                        if flatten_output:
                            output = tf.reshape(output, (-1, output.shape[-1]))
                        if one_hot_target:
                            target = tf.one_hot(
                                tf.squeeze(target), depth=output.shape[-1]
                            )

                    if lossfn == "xent":
                        loss = tf.keras.losses.CategoricalCrossentropy(
                            from_logits=True
                        )(target, output)
                    elif lossfn == "mse":
                        loss = tf.keras.losses.MeanSquaredError()(target, output)
                    elif lossfn == "poisson":
                        loss = tf.keras.losses.Poisson()(target, output)
                    else:
                        raise NotImplementedError(f"unknown `lossfn`: {lossfn}")

                    # compute gradients
                    gradients = tape.gradient(loss, model.trainable_variables)

                    df.extend(
                        get_input_output(
                            model, data, t, width, skip_tensors=skip_tensors
                        )
                    )

                    optimizer.apply_gradients(
                        (grad, var)
                        for (grad, var) in zip(gradients, model.trainable_variables)
                        if grad is not None
                    )
                    del gradients

                if t == nsteps:
                    break

            del model

            if show_progress:
                pbar.update(1)
    if show_progress:
        pbar.close()
    return pd.DataFrame(df)


def get_coord_data(
    models,
    dataloader,
    optimizer="sgd",
    learning_rate=None,
    mup=True,
    filter_trainable_by_name=None,
    skip_tensors=None,
    **kwargs,
):
    """Get coord data for coord check.

    Train the models in `models` with data from `dataloader` and optimizer
    specified by `optimizer` and `lr` for `nsteps` steps, and record coordinate
    statistics specified by `output_fdict`, `input_fdict`, `param_fdict`. By
    default, only `l1` is computed for output activations of each module.

    This function wraps around `_get_coord_data`, with the main difference being
    user can specify common optimizers via a more convenient interface.

    Args:
        models:
            a dict of lazy models, where the keys are numbers indicating width.
            Each entry of `models` is a function that instantiates a model given
            nothing.
        dataloader:
            an iterator whose elements are either Huggingface style dicts, if
            `dict_in_out` is True, or (input, label). If `fix_data` is True
            (which is the default), then only the first element of `dataloader`
            is used in a loop and the rest of `dataloder` is ignored.
        optimizer:
            a string in `['sgd', 'adam', 'adamw']`, with default being `'sgd'`.
        learning_rate:
            learning rate. By default is 0.1 for `'sgd'` and 1e-3 for others.
        mup:
            If True, then use the optimizer from `mup.optim`; otherwise, use the
            one from `torch.optim`.
        filter_trainable_by_name:
            a function that returns a bool given module names (from
            `model.layers`), or None. If not None, then only modules
            whose name yields True will be trained.
        nsteps:
            number of steps to train the model
        dict_in_out:
            whether the data loader contains Huggingface-style dict input and
            output. Default: False
        flatten_input:
            if not `dict_in_out`, reshape the input to be
            `input.view(input.shape[0], -1)`. Typically used for testing MLPs.
        flatten_output:
            if not `dict_in_out`, reshape the label to be `label.view(-1,
            input.shape[-1])`.
        output_name:
            if `dict_in_out`, this is the key for the loss value if the output
            is a dict. If the output is not a dict, then we assume the first
            element of the output is the loss.
        lossfn:
            loss function to use if not `dict_in_out`. Can be either a string from
            [`xent`, 'mse']. Default: 'xent'
        filter_module_by_name:
            a function that returns a bool given module names (from
            `model.layers`), or None. If not None, then only modules
            whose name yields True will be recorded.
        nseeds:
            number of times to repeat the training, each with different seeds.
        output_fdict, input_fdict, param_fdict:
            function dicts to be used in `_record_coords`. By default, only `l1`
            is computed for output activations of each module.
        show_progress:
            show progress using tqdm. Default: True
        one_hot_target:
            convert target label into a one-hot vector. This typically is only
            used for `'mse'` or `'l1'` losses in classification tasks.
            Default: False
    Returns:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'module', 't'` as well as names of statistics recorded, such
        as `'l1'` (see `FDICT` for other premade statistics that can be
        collected).
    """
    if learning_rate is None:
        learning_rate = 0.1 if optimizer == "sgd" else 1e-3
    if mup:
        from mup_tf import MuAdam as Adam
        from mup_tf import MuSGD as SGD
    else:
        from tensorflow.keras.optimizers import SGD, Adam

    def get_trainable(model):
        params = model.trainable_weights
        if filter_trainable_by_name is not None:
            params = []
            for tensor in model.trainable_weights:
                if filter_trainable_by_name(tensor.name):
                    params.append(tensor)
        return params

    if optimizer == "sgd":
        optimizer = SGD(learning_rate=learning_rate)
    elif optimizer == "adam":
        optimizer = Adam(learning_rate=learning_rate)
    elif optimizer is None:
        raise ValueError("optimizer should be sgd|adam")

    data = _get_coord_data(
        models, dataloader, optimizer, skip_tensors=skip_tensors, **kwargs
    )
    data["optimizer"] = optimizer
    data["learning_rate"] = learning_rate
    return data


def plot_coord_data(
    df,
    y="l1",
    save_to=None,
    suptitle=None,
    x="width",
    hue="module",
    legend="auto",
    name_contains=None,
    name_not_contains=None,
    loglog=True,
    logbase=2,
    face_color=None,
    subplot_width=20,
    subplot_height=16,
):
    """Plot coord check data `df` obtained from `get_coord_data`.

    Args:
        df:
            a pandas DataFrame obtained from `get_coord_data`
        y:
            the column of `df` to plot on the y-axis. Default: `'l1'`
        save_to:
            path to save the resulting figure, or None. Default: None.
        suptitle:
            The title of the entire figure.
        x:
            the column of `df` to plot on the x-axis. Default: `'width'`
        hue:
            the column of `df` to represent as color. Default: `'module'`
        legend:
            'auto', 'brief', 'full', or False. This is passed to `seaborn.lineplot`.
        name_contains:
            only plot modules whose name contains `name_contains`
        name_not_contains:
            only plot modules whose name does not contain `name_not_contains`
        loglog:
            whether to use loglog scale. Default: True
        logbase:
            the log base, if using loglog scale. Default: 2
        face_color:
            background color of the plot. Default: None (which means white)
        subplot_width, subplot_height:
            The width and height for each timestep's subplot. More precisely,
            the figure size will be
                `(subplot_width*number_of_time_steps, subplot_height)`.
            Default: 5, 4

    Returns:
        the `matplotlib` figure object
    """
    ### preprocessing
    df = copy(df)
    df = df[df.module != ""]

    if name_contains is not None:
        df = df[df["module"].str.contains(name_contains)]
    elif name_not_contains is not None:
        df = df[~(df["module"].str.contains(name_not_contains))]

    ts = df.t.unique()

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set()

    ### plot
    fig = plt.figure(figsize=(subplot_width * len(ts), subplot_height))
    if face_color is not None:
        fig.patch.set_facecolor(face_color)
    for t in ts:
        plt.subplot(1, len(ts), t)
        sns.lineplot(
            x=x, y=y, data=df[df.t == t], hue=hue, legend=legend if t == 1 else None
        )
        plt.title(f"t={t}")
        if t != 1:
            plt.ylabel("")
        if loglog:
            plt.loglog(base=logbase)
    if suptitle:
        plt.suptitle(suptitle)
    if save_to is not None:
        plt.savefig(save_to)
        print(f"coord check plot saved to {save_to}")

    return fig



from setuptools import find_packages, setup

packages = find_packages(exclude=("setup", "tests"))

version = "0.0.2"

setup(
    name="mup_tf",
    version=version,
    description="Maximal Update Parametrization in Tensorflow",
    author="Zach Nussbaum",
    author_email="zanussbaum@gmail.com",
    url="https://github.com/zanussbaum/mup-tf",
    packages=packages,
    include_package_data=True,  # for manifest files.
    python_requires=">=3.7, <4",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "tensorflow>=2.6.2,<2.10.0",
        "tqdm",
        "pyyaml",
        "matplotlib",
        "seaborn",
    ],
    license="PROPRIETARY",
    test_suite="tests",
    tests_require=["pytest-runner"],
)



import re
from copy import deepcopy

import tensorflow as tf
import yaml

from mup_tf.infshape import InfShape, zip_infshape
from mup_tf.layer import MuReadout, extract_layers, reinit_layer

__BSH_COMMENT__ = """\
# This is a base shape file encoded in yaml
# - `null` indicates a dimension is "finite", i.e. a non-"width" dimension
# - a number indicates the base dimension of an "infinite" dimension, i.e. some notion of "width"
"""


def get_base_name(name):
    """Get the base name of a layer. e.g. returns dense from dense

    Args:
        name: name of a layer

    Returns:
        base name of a layer
    """
    return re.sub(r"_[\d]+", "", name)


def count_layers(shapes):
    """Count the number of layers in a dict of shapes.

    Args:
        shapes: an iterable of shapes

    Returns:
        dict of base layer name to number of layers
    """
    counter = {}
    for name in shapes:
        split = get_base_name(name)
        count = counter.get(split, 0)
        counter[split] = count + 1

    return counter


def rename_tensor_names(shapes, tensors):
    """Rename the tensor names in a dict of shapes to be locally named.

    Tensorflow names tensors with a global name. For example, imagine you have
    two models each with a dense layer. The first dense layer in the first model
    would be named `dense/kernel:0` and the first dense layer in the second model
    woudl be named `dense_1/kernel:0`. This function renames the tensors to be
    locally named so each tensor is named `dense/kernel:0` in both models. This
    is needed so that we can map layers across models.

    Args:
        shapes: a dict of shapes where the keys are tensor names and the values
            are the shapes of the tensors
        tensors: a list of tensors.

    Returns:
        a dict of shapes where the keys are locally named tensor names

    """
    layer_name_count = count_layers(shapes)
    counted_tensors = {key: 0 for key in layer_name_count}

    renamed_tensors = {}
    for tensor in tensors:
        base_name = get_base_name(tensor.name)
        shape = shapes[tensor.name]
        if base_name not in counted_tensors:
            counted_tensors[base_name] = 0
        else:
            counted_tensors[base_name] += 1
        new_name = f"{base_name}_layer_{counted_tensors[base_name]}"
        renamed_tensors[new_name] = shape

        tensor.mapped_name = new_name

    assert len(renamed_tensors) == len(tensors)
    assert list(renamed_tensors.values()) == list(shapes.values())

    return renamed_tensors


def get_shapes(model):
    """Gets the shapes of each weight in a model.

    Args:
        model: a `tf.keras.Model`

    Returns:
        a dict of shapes where the keys are the names of the weights and the values
        are the shapes of the weights
    """
    tensor2shape = {tensor.name: tensor.shape for tensor in model.trainable_weights}

    tensor2shape = rename_tensor_names(tensor2shape, model.trainable_weights)
    return tensor2shape


def get_infshapes(model):
    """Gets the `mup_tf.InfShape` of each weight in a model.

    Args:
        model: a `tf.keras.Model`

    Returns:
        a dict of shapes where the keys are the names of the weights and the values
        are the `mup_tf.InfShape` of the weights
    """
    tensor_names = [tensor.name for tensor in model.trainable_weights]
    tensor2infshape = {
        tensor.name: tensor.infshape for tensor in model.trainable_weights
    }
    tensor2infshape = rename_tensor_names(tensor2infshape, tensor_names)
    return tensor2infshape


def save_base_shapes(model_or_shapes, file):
    """Saves the shapes of a model to a file.

    Args:
        model_or_shapes: a `tf.keras.Model` or a dict of shapes
        file: a file to save the shapes to

    Raises:
        ValueError: if `model_or_shapes` is not a `tf.keras.Model` or a dict of shapes
    """
    if isinstance(model_or_shapes, tf.keras.layers.Layer):
        sh = get_infshapes(model_or_shapes)
    elif isinstance(model_or_shapes, dict):
        sh = deepcopy(model_or_shapes)
    else:
        raise ValueError()
    sh = {k: s.base_shape() for k, s in sh.items()}
    s = yaml.dump(sh, None, indent=4)
    s = __BSH_COMMENT__ + s
    with open(file, "w") as f:
        f.write(s)


def load_base_shapes(filename):
    """Get a dict of `InfShape` from a filename.

    Args:
        filename: a filename to load the shapes from

    Returns:
        a dict of `InfShape` where the keys are the names of the weights and the values
    """
    with open(filename, "r") as f:
        d = yaml.safe_load(f)
    return {k: InfShape.from_base_shape(v) for k, v in d.items()}


def _extract_shapes(x):
    """
    Args:
        x: can be any of the following:
            - `tf.keras.layers.Layer`
            - dict of shapes
            - dict of `InfShape`
            - str of path to a base shapes (.bsh) file
    Returns:
        If `x` is dict of `InfShape`, then output itself.
        If `x` is path, then output a dict of `InfShapes` loaded from `x`.
        Else, output the shapes (not `InfShape`) associated to `x`
    """
    if isinstance(x, tf.keras.layers.Layer):
        x_shapes = get_shapes(x)
    elif isinstance(x, dict):
        x_shapes = deepcopy(x)
    elif isinstance(x, str):
        # x is file name
        x_shapes = load_base_shapes(x)
    else:
        raise ValueError(f"unhandled x type: {type(x)}")
    return x_shapes


def assert_equal_shapes(base_shapes, shapes):
    """Assert that two dicts of shapes are equal.

    Args:
        base_shapes: a dict of shapes
        shapes: a dict of shapes
    """
    base_layers_counter = count_layers(base_shapes)

    layers_counter = count_layers(shapes)

    if layers_counter != base_layers_counter:
        raise ValueError(
            f"layers_counter != base_layers_counter: {layers_counter} != {base_layers_counter}"
        )


def _zip_infshape_dict(base_shapes, shapes):
    """Make a dict of `InfShape` from two dicts of shapes.

    Args:
        base_shapes: dict of base shapes or InfShape objects
        shapes: dict of shapes
    Returns:
        dict of `InfShape` using `zip_infshape`
    """
    assert_equal_shapes(base_shapes, shapes)

    infshapes = {}
    for name, bsh in base_shapes.items():
        infshapes[name] = zip_infshape(bsh, shapes[name])
    return infshapes


def zip_infshapes(base, target):
    """Make a dict of `InfShape` from models or dicts.

    Args:
        base: a base `tf.keras.layers.Layer` or a dict of shapes
        target: a target `tf.keras.layers.Layer` or a dict of shapes
    Returns:
        dict of `InfShape` using `zip_infshape`
    """
    base_shapes = _extract_shapes(base)
    target_shapes = _extract_shapes(target)
    return _zip_infshape_dict(base_shapes, target_shapes)


def clear_dims(infshape_dict):
    """
    Args:
        infshape_dict: dict of `InfShape`
    Returns:
        the same dict but where all `InfDim` in all `InfShape`
        have their `dim` attribute set to None
    """
    d = deepcopy(infshape_dict)
    for _, v in d.items():
        for infdim in v:
            infdim.dim = None
    return d


def make_base_shapes(base_shapes, delta_shapes, savefile=None):
    """Make a base shape object from a base model/shapes and a delta model/shapes.

    Args:
        base:
            a base `tf.keras.models.Model` or a dict of shapes
        delta:
            a "delta" model or a dict of shapes, for the sole purpose of
            determining which dimensions are "width" and will be scaled up and
            down in the target model.
        savefile:
            if a string, then the resulting base shape object is serialized to
            this location via yaml encoding.
    Returns:
        base infshapes
    """
    bsh = clear_dims(zip_infshapes(base_shapes, delta_shapes))
    if savefile is not None:
        save_base_shapes(bsh, savefile)
    return bsh


def apply_infshapes(model, infshapes):
    """Apply a dict of `InfShape` to a model.

    Args:
        model: a `tf.keras.models.Model`
        infshapes: a dict of tensor_name: `InfShape`
    """
    for tensor in model.trainable_weights:
        tensor.infshape = infshapes[tensor.mapped_name]


def set_base_shapes(
    model, base, rescale_params=True, delta=None, savefile=None, do_assert=True
):
    """Sets the `p.infshape` attribute for each parameter `p` of `model`.

    Args:
        model: tf.keras.models.Model instance
        base: The base model.
            Can be tf.keras.models.Model, a dict of shapes, a str, or None.
            If None, then defaults to `model`
            If str, then treated as filename for yaml encoding of a dict of base shapes.
        rescale_params:
            assuming the model is initialized using the default init (or
            He initialization etc that scale the same way with fanin): If True
            (default), rescales parameters to have the correct (μP) variances.
        do_assert:
    Returns:
        same object as `model`, after setting the `infshape` attribute of each parameter.
    """
    if not model.built:
        raise ValueError("model must be built before setting base shapes")

    if base is None:
        base = model
    base_shapes = _extract_shapes(base)
    if delta is not None:
        delta_shapes = _extract_shapes(delta)
        base_shapes = _zip_infshape_dict(base_shapes, delta_shapes)
    shapes = get_shapes(model)
    infshapes = _zip_infshape_dict(base_shapes, shapes)
    if savefile is not None:
        save_base_shapes(infshapes, savefile)
    apply_infshapes(model, infshapes)
    if do_assert:
        assert_hidden_size_inf(model)
    if rescale_params:
        layers = extract_layers(model)
        for layer in layers:
            if isinstance(layer, MuReadout):
                layer._rescale_parameters()
            elif isinstance(
                layer,
                (tf.keras.layers.Dense, tf.keras.layers.Conv2D, tf.keras.layers.Conv1D),
            ):
                reinit_layer(layer)

    name2shape = {k.name: infshapes[k.mapped_name] for k in model.trainable_weights}
    return name2shape


def assert_hidden_size_inf(model):
    """
    This tests for any `Dense` whose output dimension is finite but input
    dimension is infinite and is not of type `MuReadout`. Such `Dense`
    modules should not exist in a correctly parametrized models.

    Args:
        model: a `tf.keras.models.Model`
    """
    for tensor in model.trainable_weights:
        if isinstance(tensor, tf.keras.layers.Dense) and not isinstance(
            tensor, MuReadout
        ):
            if not tensor.infshape[0].isinf() and tensor.infshape[1].isinf():
                raise ValueError(
                    f"{tensor.mapped_name} has infinite fan-in and finite fan-out dimensions but is not type `MuReadout`. "
                    "To resolve this, either change the module to `MuReadout` or change the fan-out to an infinite dimension."
                )
